#include <cmath>
#include <cstring>
#include <cstdio>
#include <algorithm>
#include <unistd.h>

#include "resample.hh"
#include "argh.hh"

#define MagickPI M_PI
#include "filters.inc"

static const struct
{
    double (*function)(double);
    double support;
    const char *name;
} Filters[] = 
{
    { Box,      0.0, "none"      },
    { Box,      0.5, "linear"    },
    { Triangle, 1.0, "triangle"  },
    { Hermite,  1.0, "hermite"   },
    { Hanning,  1.0, "hanning"   },
    { Hamming,  1.0, "hamming"   },
    { Blackman, 1.0, "blackman"  },
    { Gaussian, 1.25, "gaussian" },
    { Quadratic,1.5, "quadratic" },
    { Cubic,    2.0, "cubic"     },
    { Catrom,   2.0, "catrom"    },
    { Mitchell, 2.0, "mitchell"  },
    { Lanczos,  3.0, "lanczos"   },
    { Bessel, 3.2383, "bessel"   },
    { Sinc,     4.0, "sinc"      }
};
static const unsigned FilterCount = (sizeof Filters) / (sizeof(Filters[0]));
static unsigned Filter = 1;  /* linear */
static double blur = 1;      /* <1 = sharp, >1 = blurry */

const stent &Stream::operator[] (int index)
{
    while(index >= basepos + inbufsize)
    {
        basepos += leapstep;
        memmove(&inent[0], &inent[leapstep], min(leapstep,newsize) * sizeof(inent[0]));
        if(!Read(&inent[leapstep], newsize))
        {
            /* EOF */
            exit(0);
        }
    }
    if(index < basepos)return inent[0];
    return inent[index - basepos];
}

static void Interpolate(Stream &input, Stream &output)
{
    if(input.rate() == output.rate())
    {
        // No change in rate, just copy
        stent tab[8192];
        for(;;)
        {
            unsigned oco = input.Read(tab, 8192);
            if(!oco)return;
            output.Write(tab, 8192);
        }
    }
    
    double factor = (double)input.rate() / output.rate();
    
    double scale_factor = blur * max(factor, 1.0);
    double support = max(scale_factor * Filters[Filter].support, 0.5);
    if(support <= 0.5)
    {
        /* reduce to point sampling (??) */
        support = 0.5 + 1E-12;
        scale_factor = 1.0;
    }
    
    double weight[(int)support*2+3];    
    double (*const filterfun)(double) = Filters[Filter].function;
    
    double center = 0.0;
    const unsigned outbufsize = output.rate() / 8;
    stent oent[outbufsize];

    for(;;)
    {
        for(unsigned o=0; o<outbufsize; ++o, center += factor)
        {
            int start = (int)(0.5 + center - support);
            int end   = (int)(0.5 + center + support);
            
            register double density = 0;
            int n = 0;
            for(int i = start; i < end; ++i)
                density += (weight[n++] = filterfun((i-center+0.5) / scale_factor) / scale_factor);
            
            if(density != 0.0)
                for(register int i=0; i<n; ++i)
                    weight[i] /= density; /* normalize */
            
            register double l=0.0;
            register double r=0.0;
            for(register int i=0, p=start; i<n; ++i)
            {
                const stent &srcent = input[p++];
                l += weight[i] * srcent.l;
                r += weight[i] * srcent.r;
            }
            oent[o].l = l<-32768 ? -32768 : l>32767 ? 32767 : (int)(l+0.5);
            oent[o].r = r<-32768 ? -32768 : r>32767 ? 32767 : (int)(r+0.5);
        }

        output.Write(oent, outbufsize);
        
        register int shiftval = (int)center;
        /* Prevent overflowing in 17 hours */
        input.shift(shiftval);
        center -= shiftval;
    }
}

Stream::Stream(const char *settings) : bit16(true), stereo(true),
                                       herz(0), basepos(-inbufsize)
{
    while(*settings)
    {
        if(*settings=='m')stereo = false;
        else if(*settings=='s')stereo = true;
        else if(*settings=='b')bit16 = false;
        else if(*settings=='w')bit16 = true;
        else if(*settings=='r')herz = 0;
        else if(*settings>='0' && *settings<='9')
            herz = herz * 10 + (*settings - '0');
        else if(*settings!='-')throw *settings;
        ++settings;
    }
    if(!herz)herz = 44100;
}

unsigned Stream::Read(stent *dest, unsigned count) const
{
    return bit16?stereo?SafeRead(dest, count*4)/4:read16(dest,count)
                :stereo?readste(dest,count):read8(dest,count);
}

unsigned Stream::read16(stent *dest, unsigned count)
{
    int16 tab[count];
    count = SafeRead(tab, sizeof tab) / 2;
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a];
    }
    return count;
}
unsigned Stream::readste(stent *dest, unsigned count)
{
    uint8 tab[count*2];
    count = SafeRead(tab, sizeof tab) / 2;
    for(unsigned a=0, b=0; b<count; ++b)
    {
        dest[b].r = tab[a++]*256 - 32768;
        dest[b].l = tab[a++]*256 - 32768;
    }
    return count;
}
unsigned Stream::read8(stent *dest, unsigned count)
{
    uint8 tab[count];
    count = SafeRead(tab, sizeof tab);
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a]*256 - 32768;
    }
    return count;
}
void Stream::Write(const stent *dest, unsigned count) const
{
    if(bit16)if(stereo)SafeWrite(dest,count*4);else write16(dest,count);
    else if(stereo)writeste(dest,count);else write8(dest,count);
}

unsigned Stream::SafeRead(void *d, unsigned bcount)
{
    uint8 *dd = (uint8 *)d;
    unsigned r = 0;
    while(bcount > 0)
    {
        int res = std::read(0, dd, bcount);
        if(res <= 0)break;
        bcount -= res;
        r += res;
        dd += res;
    }
    return r;
}
void Stream::SafeWrite(const void *d, unsigned bcount)
{
    std::write(1, d, bcount);
}
void Stream::write16(const stent *src, unsigned count)
{
    int16 tab[count];
    for(unsigned a=0; a<count; ++a)
    {
        tab[a] = (src[a].r + src[a].l) / 2;
    }
    SafeWrite(tab, sizeof tab);
}
void Stream::writeste(const stent *src, unsigned count)
{
    uint8 tab[count*2];
    for(unsigned a=0, b=0; b<count; ++b)
    {
        tab[a++] = (src[b].r + 32768) / 256;
        tab[a++] = (src[b].l + 32768) / 256;
    }
    SafeWrite(tab, sizeof tab);
}
void Stream::write8(const stent *src, unsigned count)
{
    uint8 tab[count];
    for(unsigned a=0; a<count; ++a)
    {
        tab[a] = ((src[a].r + src[a].l) / 2 + 32768) / 256;
    }
    SafeWrite(tab, sizeof tab);
}

#define VERSIONSTR \
    "resample v"VERSION" - Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "Portions copyright (C) 2001 ImageMagick Studio.\n" \
    "This program is under GPL. wavegenerator-"VERSION".{rar,zip,tar.{gz,bz2}}\n" \
    "are available at the homepage of the author.\n"

class Handle : public arghandler
{
    bool Help;
public:
    string opt_h(const string &s)
    {
        Help = true;
        return s;
    }
    string opt_V(const string &)
    {
        printf(VERSIONSTR);
        exit(0);
    }
    string opt_b(const string &s)
    {
        blur = 0;
        unsigned a=0;
        while(a<s.size() && isdigit(s[a]))blur = blur*10 + s[a++] - '0';
        blur = 1+blur;
        return s.substr(a);
    }
    string opt_s(const string &s)
    {
        blur = 0;
        unsigned a=0;
        while(a<s.size() && isdigit(s[a]))blur = blur*10 + s[a++] - '0';
        blur = 1 - blur;
        return s.substr(a);
    }
    string opt_f(const string &s)
    {
        for(unsigned a=0; a<FilterCount; ++a)
            if(Filters[a].name == s)
            {
                Filter = a;
                return "";
            }
        fprintf(stderr, "'%s' is not a supported filter.\n", s.c_str());
        exit(-1);
    }
    Handle(const char *defopts, int argc, const char *const *argv)
    : arghandler(defopts, argc, argv),
        Help(false)
    {
        add("-h", "--help",     "This text",    &(Handle::opt_h));
        
        add("-V", "--version",  "Version info", &(Handle::opt_V));
        
        add("-b", "--blur",     "blur: 0=normal, >0=blur\n"
                                "(note: blurring requires cpu power.)",
                                &(Handle::opt_b));
        
        add("-s", "--sharpen",  "sharpen: 0=normal, >0=sharper\n"
                                "(note: blurring and sharping are exclusive.)",
                                &(Handle::opt_s));
        
        string FiltText = "Select filter of the following:\n";
        for(unsigned p=0, a=0; a<FilterCount; ++a)
        {
            string L = Filters[a].name;
            if(a==Filter)L += " (default)";
            if(p + L.size() > 30) { p = 0; FiltText += "\n"; } else if(a > 0)FiltText += ' ';
            if(!p)FiltText += "   ";
            FiltText += L;
            p += L.size();
            if(a+1 < FilterCount)FiltText += ',';
        }
        add("-f", "--filter",   FiltText, &(Handle::opt_f));
    }
    virtual void defarg(const string &)
    {
    }
    void Usage()
    {
        printf(
            "Usage: %s <inputspec> <outputspec> [<options> ...]\n"
            "\n"
            "Reads stdin, writes stdout.\n"
            "\n"
            "input- and outputspec:\n"
            "     mr22050 = mono 22050 Hz 16-bit\n"
            "     b       = 8-bit 44100 Hz stereo\n"
            "     mb      = 8-bit 44100 Hz mono\n"
            "     br22050 = 8-bit 22050 Hz stereo\n"
            "     m       = mono 44100 Hz 16-bit\n"
            "   etc. - optional.\n"
            "\nOptions:\n", a0.c_str());
        
        listoptions();
        
        exit(0);
    }
    virtual void parse()
    {
        arghandler::parse();
        if(Help)Usage();
    }
};

int main(int argc, const char *const *argv)
{
    Handle parameters("", argc > 2 ? (2-argc)*MAX_ARGC-2 : argc, argv);
    if(argc <= 2)
    {
        parameters.parse();
        parameters.Usage();
    }
    
    Stream *input, *output;
    
    try
    {
        input = new Stream(argv[1]);
        output = new Stream(argv[2]);
    }
    catch(char c)
    {
        fprintf(stderr, "Invalid option: '%c'\n", c);
        return -1;
    }
    
    parameters.parse(); 
    Interpolate(*input, *output);
    return 0;
}
